"""Object caching"""

import os
from butcher import gitrepo
from butcher import util
from twitter.common import app
from twitter.common import log

app.add_option('--object_cache_dir', dest='object_cache_dir',
               help='Path to local object cache directory.')


class CacheMiss(RuntimeError):
    pass


class CacheManager(app.Module):
    """Built object cache manager."""
    cache_dir = None

    def __init__(self):
        app.Module.__init__(self, label='butcher.cachemanager',
                            description='Butcher cache manager')
        self.obj_cachedir = None
        self.mh_cachedir = None

    def setup_function(self):
        if not self.cache_dir:
            opts = app.get_options()
            if opts.object_cache_dir:
                self.cache_dir = opts.object_cache_dir
            else:
                self.cache_dir = os.path.join(opts.butcher_basedir, 'cache')
        self.obj_cachedir = os.path.join(self.cache_dir, 'obj')
        self.mh_cachedir = os.path.join(self.cache_dir, 'mh')
        for subdir in (self.obj_cachedir, self.mh_cachedir):
            if not os.path.exists(subdir):
                os.makedirs(subdir)
        log.info('Cache initialized at %s', self.cache_dir)

    def path_in_cache(self, filename, metahash):
        """Generates the path to a file in the mh cache.

        The generated path does not imply the file's existence!

        Args:
          filename: Filename relative to buildroot
          rule: A targets.SomeBuildRule object
          metahash: hash object
        """
        cpath = self._genpath(filename, metahash)
        if os.path.exists(cpath):
            return cpath
        else:
            raise CacheMiss

    def _genpath(self, filename, mhash):
        """Generate the path to a file in the cache.

        Does not check to see if the file exists. Just constructs the path
        where it should be.
        """
        mhash = mhash.hexdigest()
        return os.path.join(self.mh_cachedir, mhash[0:2], mhash[2:4],
                            mhash, filename)

    def putfile(self, filepath, buildroot, metahash):
        """Put a file in the cache.

        Args:
          filepath: Path to file on disk.
          buildroot: Path to buildroot
          buildrule: The rule that generated this file.
          metahash: hash object
        """
        def gen_obj_path(filename):
            filehash = util.hash_file(filepath).hexdigest()
            return filehash, os.path.join(self.obj_cachedir, filehash[0:2],
                                          filehash[2:4], filehash)

        filepath_relative = filepath.split(buildroot)[1][1:]  # Strip leading /
        # Path for the metahashed reference:
        incachepath = self._genpath(filepath_relative, metahash)

        filehash, obj_path = gen_obj_path(filepath)
        if not os.path.exists(obj_path):
            obj_dir = os.path.dirname(obj_path)
            if not os.path.exists(obj_dir):
                os.makedirs(obj_dir)
            log.debug('Adding to obj cache: %s -> %s', filepath, obj_path)
            os.link(filepath, obj_path)

        if os.path.exists(incachepath):
            existingfile_hash = util.hash_file(incachepath).hexdigest()
            if filehash != existingfile_hash:
                log.warn('File found in mh cache, but checksum differs. '
                         'Replacing with this new version. (File: %s)',
                         filepath)
                log.warn('Possible reasons for this:')
                log.warn(' 1. This build is not hermetic, and something '
                         'differs about the build environment compared to the '
                         'previous build.')
                log.warn(' 2. This file has a timestamp or other build-time '
                         'related data encoded into it, which will always '
                         'cause the checksum to differ when built.')
                log.warn(' 3. Everything is terrible and nothing works.')
                os.unlink(incachepath)

        if not os.path.exists(incachepath):
            log.debug('Adding to mh cache: %s -> %s', filepath, incachepath)
            if not os.path.exists(os.path.dirname(incachepath)):
                os.makedirs(os.path.dirname(incachepath))
            os.link(obj_path, incachepath)

    def in_cache(self, objpath, metahash):
        """Returns true if object is cached.

        Args:
          objpath: Filename relative to buildroot.
          metahash: hash object
        """
        try:
            self.path_in_cache(objpath, metahash)
            return True
        except CacheMiss:
            return False

    def get_obj(self, objpath, metahash, dst_path):
        """Get object from cache, write it to dst_path.

        Args:
          objpath: filename relative to buildroot
                   (example: mini-boot/blahblah/somefile.bin)
          metahash: metahash. See targets/base.py
          dst_path: Absolute path where the file should be written.
        Raises:
          CacheMiss: if the item is not in the cache
        """
        incachepath = self.path_in_cache(objpath, metahash)
        if not os.path.exists(incachepath):
            raise CacheMiss('%s not in cache.' % incachepath)
        else:
            log.debug('Cache hit! %s~%s', objpath, metahash.hexdigest())
            if not os.path.exists(os.path.dirname(dst_path)):
                os.makedirs(os.path.dirname(dst_path))
            os.link(incachepath, dst_path)
            #shutil.copy2(incachepath, dst_path)



"""Base target."""

import os
import re
import shutil
from butcher import address
from butcher import cache
from butcher import error
from butcher import util
from twitter.common import log


class BaseBuilder(object):
    """Base class for rule self-builders."""

    # If true, butcher will generally hard-link files from its cache into the
    # build area rather than copying them.
    linkfiles = True

    def __init__(self, buildroot, target_obj, source_dir):
        self.source_dir = source_dir    # Where the git repo is checked out
        self.rule = target_obj          # targets.something object
        self.address = target_obj.name  # Build address
        self.buildroot = buildroot
        self.cachemgr = cache.CacheManager()
        self._cached_metahash = None
        if not os.path.exists(self.buildroot):
            os.makedirs(self.buildroot)
        # TODO: some rule types don't have srcs.
        #       Should probably use an intermediate subclass.
        self.srcs_map = {}
        self.deps_map = {}

    def collect_srcs(self):
        for src in self.rule.source_files or []:
            srcpath = os.path.join(self.source_dir, self.address.path, src)
            dstpath = os.path.join(self.buildroot, self.address.repo,
                                   self.address.path, src)
            dstdir = os.path.dirname(dstpath)
            if not os.path.exists(dstdir):
                os.makedirs(dstdir)
            log.debug('[%s]: Collect srcs: %s -> %s', self.rule.address,
                      srcpath, dstpath)
            self.linkorcopy(srcpath, dstpath)
            self.srcs_map[src] = dstpath

    def collect_deps(self):
        pass
        # Holy crap, this whole function is unnecessary with the current
        # design. A rule's dependencies are built before the rule (duh) and the
        # outputs thereof go into the same buildroot that this one uses, so the
        # files will already be in place. This will probably not always be the
        # case though, so this ought to get implemented in an intelligent way.

        #if 'deps' not in self.rule.params:
        #    return
        #for dep in self.rule.composed_deps() or []:
        #    dep_rule = self.rule.subgraph.node[dep]['target_obj']
        #    for item in dep_rule.output_files:
        #       srcpath = os.path.join(self.buildroot, item)
        #       dstpath = os.path.join(self.buildroot, item)
        #       dstdir = os.path.dirname(dstpath)
        #       log.debug('[%s]: Collecting deps: %s -> %s',
        #                 self.address, item, dstpath)
        #       if not os.path.exists(dstdir):
        #           os.makedirs(dstdir)
        #       shutil.copy2(srcpath, dstdir)
        #       self.deps_map[item] = dstpath

    def _metahash(self):
        """Checksum hash of all the inputs to this rule.

        Output is invalid until collect_srcs and collect_deps have been run.

        In theory, if this hash doesn't change, the outputs won't change
        either, which makes it useful for caching.
        """

        # BE CAREFUL when overriding/extending this method. You want to copy
        # the if(cached)/return(cached) part, then call this method, then at
        # the end update the cached metahash. Just like this code, basically,
        # only you call the method from the base class in the middle of it. If
        # you get this wrong it could result in butcher not noticing changed
        # inputs between runs, which could cause really nasty problems.
        # TODO(ben): the above warning seems avoidable with better memoization

        if self._cached_metahash:
            return self._cached_metahash

        # If you are extending this function in a subclass,
        # here is where you do:
        # BaseBuilder._metahash(self)

        log.debug('[%s]: Metahash input: %s', self.address,
                  unicode(self.address))
        mhash = util.hash_str(unicode(self.address))
        log.debug('[%s]: Metahash input: %s', self.address, self.rule.params)
        mhash = util.hash_str(str(self.rule.params), hasher=mhash)
        for src in self.rule.source_files or []:
            log.debug('[%s]: Metahash input: %s', self.address, src)
            mhash = util.hash_str(src, hasher=mhash)
            mhash = util.hash_file(self.srcs_map[src], hasher=mhash)
        for dep in self.rule.composed_deps() or []:
            dep_rule = self.rule.subgraph.node[dep]['target_obj']
            for item in dep_rule.output_files:
                log.debug('[%s]: Metahash input: %s', self.address, item)
                item_path = os.path.join(self.buildroot, item)
                mhash = util.hash_str(item, hasher=mhash)
                mhash = util.hash_file(item_path, hasher=mhash)
        self._cached_metahash = mhash
        return mhash

    def collect_outs(self):
        """Collect and store the outputs from this rule."""
        # TODO: this should probably live in CacheManager.
        for outfile in self.rule.output_files or []:
            outfile_built = os.path.join(self.buildroot, outfile)
            if not os.path.exists(outfile_built):
                raise error.TargetBuildFailed(
                    self.address, 'Output file is missing: %s' % outfile)

        #git_sha = gitrepo.RepoState().GetRepo(self.address.repo).repo.commit()
        # git_sha is insufficient, and is actually not all that useful.
        # More factors to include in hash:
        # - commit/state of source repo of all dependencies
        #   (or all input files?)
        #   - Actually I like that idea: hash all the input files!
        # - versions of build tools used (?)
        metahash = self._metahash()
        log.debug('[%s]: Metahash: %s', self.address, metahash.hexdigest())
        # TODO: record git repo state and buildoptions in cachemgr
        # TODO: move cachemgr to outer controller(?)
        self.cachemgr.putfile(outfile_built, self.buildroot, metahash)

    def prep(self):
        self.collect_srcs()
        self.collect_deps()

    def is_cached(self):
        """Returns true if this rule is already cached."""
        # TODO: cache by target+hash, not per file.
        try:
            for item in self.rule.output_files:
                log.info(item)
                self.cachemgr.in_cache(item, self._metahash())
        except cache.CacheMiss:
            log.info('[%s]: Not cached.', self.address)
            return False
        else:
            log.info('[%s]: found in cache.', self.address)
            return True

    def get_from_cache(self):
        """See if this rule has already been built and cached."""
        for item in self.rule.output_files:
            dstpath = os.path.join(self.buildroot, item)
            self.linkorcopy(
                self.cachemgr.path_in_cache(item, self._metahash()),
                dstpath)
            #self.cachemgr.get_obj(item, self._metahash(), dstpath)

    def linkorcopy(self, src, dst):
        """hardlink src file to dst if possible, otherwise copy."""
        if os.path.isdir(dst):
            log.warn('linkorcopy given a directory as destination. '
                     'Use caution.')
            log.debug('src: %s  dst: %s', src, dst)
        elif os.path.exists(dst):
            os.unlink(dst)
        elif not os.path.exists(os.path.dirname(dst)):
            os.makedirs(os.path.dirname(dst))
        if self.linkfiles:
            log.debug('Linking: %s -> %s', src, dst)
            os.link(src, dst)
        else:
            log.debug('Copying: %s -> %s', src, dst)
            shutil.copy2(src, dst)

    def build(self):
        """Build the rule. Must be overridden by inheriting class."""
        raise NotImplementedError(self.rule.address)

    def rulefor(self, addr):
        """Return the rule object for an address from our deps graph."""
        return self.rule.subgraph.node[self.rule.makeaddress(addr)][
            'target_obj']


class BaseTarget(object):
    """Partially abstract base class for build targets."""
    # graphcontext is attached by PythonBuildFile. If present, instances of
    # this class add themselves to the given networkx graph.
    graphcontext = None

    rulebuilder = BaseBuilder
    ruletype = None

    # List of tuples: [('argument_name', type), ...]
    required_params = [('name', str)]

    # List of tuples: ('arg_name', type, 'default value'), ...
    # or ('arg_name', (list, of, types), 'default value'), ...
    optional_params = []

    def __init__(self, **kwargs):
        """Initialize the build rule.

        Args:
          **kwargs: Assorted parameters; see subclass implementations for
                    details.
        """
        self.address = self.name = address.new(kwargs['name'])
        # TODO: eliminate use of .name
        self.subgraph = None
        self.params = {}
        log.debug('New target: %s', self.address)

        try:
            for param_name, param_type in self.required_params:
                self.params[param_name] = kwargs.pop(param_name)
                assert isinstance(self.params[param_name], param_type)
        except AssertionError as err:
            if isinstance(param_type, tuple) and len(param_type) > 1:
                msg = 'one of: %s' % ', '.join(param_type.__name__)
            else:
                msg = str(param_type.__name__)
            raise error.InvalidRule(
                'While loading %s: Invalid type for %s. '
                'Expected: %s. Actual: %s.' % (
                    self.address, param_name, msg,
                    repr(self.params[param_name])))
        except KeyError as err:
            log.error(err)
            raise error.InvalidRule(
                'While loading %s: Required parameter %s not given.' % repr(
                    self.address, param_name))

        for (param_name, param_type, param_default) in self.optional_params:
            if param_name not in kwargs:
                self.params[param_name] = param_default
            else:
                self.params[param_name] = kwargs.pop(param_name)
                if not isinstance(self.params[param_name], param_type):
                    msg = str(param_type.__name__)
                    if isinstance(param_type, tuple) and len(param_type) > 1:
                        msg = 'one of: %s' % ', '.join(param_type.__name__)
                    raise error.InvalidRule(
                        'While loading %s: Invalid type for %s. '
                        'Expected: %s. Actual: %s.' % (
                            self.address, param_name, msg,
                            repr(self.params[param_name])))

        if kwargs:  # There are leftover arguments.
            raise error.InvalidRule(
                '[%s]: Unknown argument(s): %s' % (
                    self.address, ', '.join(kwargs.keys())))

        if self.graphcontext is not None:
            self.graphcontext.add_node(self.address, target_obj=self)
            # TODO: process deps here?

        try:
            self.validate_args()
        except AssertionError as err:
            raise error.InvalidRule('Error in %s: %s' % (self.address, err))

    def validate_args(self):
        """Input validation!"""
        def validate_name():
            allowed_re = '^[a-z](([a-z0-9_-]+)?([a-z0-9])?)?'
            assert isinstance(self.params['name'], basestring), (
                'Name must be a string, not %s' % repr(self.params['name']))
            assert re.match(allowed_re, self.params['name']), (
                'Invalid rule name: %s. Must match %s.' % (
                    repr(self.params['name']), repr(allowed_re)))
        validate_name()

        def validate_deps():
            if 'deps' in self.params:
                assert type(self.params['deps']) in (type(None), list), (
                    'Deps must be a list, not %s' % repr(self.params['deps']))
        validate_deps()

    @property
    def output_files(self):
        """Returns the list of output files from this rule.

        Should be overridden by inheriting class.
        Paths are relative to buildroot.
        """
        raise NotImplementedError(
            '[%s]: Implementation is incomplete.' % self.address)

    def composed_deps(self):
        """Dependencies of this build target."""
        if 'deps' in self.params:
            param_deps = self.params['deps'] or []
            deps = [self.makeaddress(dep) for dep in param_deps]
            return deps
        else:
            return None

    @property
    def source_files(self):
        """This rule's source files."""
        if 'srcs' in self.params and self.params['srcs'] is not None:
            return util.flatten(self.params['srcs'])

    def makeaddress(self, label):
        """Turn a label into an Address with current context.

        Adds repo and path if given a label that only has a :target part.
        """
        addr = address.new(label)
        if not addr.repo:
            addr.repo = self.address.repo
            if not addr.path:
                addr.path = self.address.path
        return addr



"""Unit tests for genrules."""
import re
import unittest

from butcher.targets import genrule
#from twitter.common import app
#from twitter.common import log


class TestGenrule(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_regexes(self):
        """Verifies that those obnoxious regexes do what they're meant to."""
        rdata = {
            genrule.GenRuleBuilder.paren_tag_re: [
                ('   $(fancystuff inside.blabla)', 'fancystuff inside.blabla'),
                ('asdf$(foobarbas)asdf', 'foobarbas'),
                ('$( location //foo/bar:bla  )', ' location //foo/bar:bla  '),
                ('$(@D)', '@D'),
                ('$(@)', '@'),
                ('$(SRCS)', 'SRCS'),
                ],
            genrule.GenRuleBuilder.noparen_tag_re: [
                ('$@', '@'),
                (' $@ ', '@'),
                ('$@Dfoo', '@Dfoo'),
                (' $@D ', '@D'),
                ('foo$bar_foo%guh bla', 'bar_foo'),
                (' $$blarg', None),
                ],
            }
        for regex in rdata:
            for (instr, outstr) in rdata[regex]:
                if outstr:
                    expected_output = [outstr]
                else:
                    expected_output = []
                match = re.findall(regex, instr)
                self.assertEqual(
                    expected_output, match,
                    '\nRegex: %s\n'
                    'Input: %s\n'
                    'Expected output: %s\n'
                    'Actual output: %s' % (repr(regex.pattern), repr(instr),
                                           expected_output, match))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestGenrule)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""Exceptions and errors."""


class ButcherError(RuntimeError):
    """Generic error class."""
    pass


class BrokenGraph(ButcherError):
    """Something's fubar in the graph. Probably bad user input."""
    pass


class InvalidRule(ButcherError):
    """That is totally not a valid build rule."""
    node = None


class BuildFailure(ButcherError):
    """Parent for build failures."""
    node = None


class AddressError(ButcherError):
    """Can't parse that or something."""
    pass


class TargetBuildFailed(BuildFailure):
    """A build failed."""
    def __init__(self, node, *args):
        BuildFailure.__init__(self, *args)
        self.node = node


class OverallBuildFailure(BuildFailure):
    """The overall build failed, most likely because of individual failures."""
    pass


class NoSuchTargetError(BuildFailure):
    """That target does not exist."""
    pass



"""Target rule types."""

from twitter.common import log
from butcher import error
from butcher.targets import gendeb
from butcher.targets import genrule
from butcher.targets import filegroup
from butcher.targets import pkgfilegroup
from butcher.targets import pkg_symlink
from butcher.targets import virtual

TYPE_MAP = {
    'genrule': genrule.GenRule,
    'gendeb': gendeb.GenDeb,
    'filegroup': filegroup.FileGroup,
    'pkgfilegroup': pkgfilegroup.PkgFileGroup,
    'virtual': virtual.VirtualTarget,
    }

__all__ = [
    'genrule',
    'gendeb',
    'filegroup',
    'pkgfilegroup',
    ]


def new(ruletype, **kwargs):
    """Instantiate a new build rule based on kwargs.

    Appropriate args list varies with rule type.
    Minimum args required:
      [... fill this in ...]
    """
    try:
        ruleclass = TYPE_MAP[ruletype]
    except KeyError:
        raise error.InvalidRule('Unrecognized rule type: %s' % ruletype)

    try:
        return ruleclass(**kwargs)
    except TypeError:
        log.error('BADNESS. ruletype: %s, data: %s', ruletype, kwargs)
        raise
        #raise error.InvalidRule(
        #    '%s does not work that way.\nDetails: %s.\nData: %s' % (
        #        ruletype, err, kwargs))



"""gendeb targets"""

import hashlib
import os
import pwd
import re
import socket
import subprocess
import sys
from butcher.targets import base
from butcher.targets import filegroup
from butcher.targets import pkgfilegroup
from butcher.targets import pkg_symlink
from butcher import error
from butcher import util
from twitter.common import app
from twitter.common import log


app.add_option(
    '--fpm_bin', dest='fpm_bin', default='fpm',
    help='Path to the fpm binary.')


class GenDebSetup(app.Module):
    """Pre-run requirements to be set up before the build process starts."""
    def __init__(self):
        app.Module.__init__(self, label=__name__,
                            description='gendeb')

    def setup_function(self):
        """twitter.comon.app runs this before any global main() function."""
        fpm_path = app.get_options().fpm_bin
        if not os.path.exists(fpm_path):
            log.warn('Could not find fpm; gendeb cannot function.')
        else:
            GenDebBuilder.fpm_bin = fpm_path

app.register_module(GenDebSetup())


class GenDebBuilder(base.BaseBuilder):
    """Builder for gendeb rules"""

    # Path to the fpm binary.
    fpm_bin = None

    def __init__(self, buildroot, target_obj, source_dir):
        base.BaseBuilder.__init__(self, buildroot, target_obj, source_dir)
        self.workdir = os.path.join(
            self.buildroot, self.address.repo, self.address.path,
            '__GENDEB_%s' % self.address.target)
        self.deb_fsroot = os.path.join(self.workdir, 'root')
        self.deb_controlroot = os.path.join(self.workdir, 'DEBIAN')
        self.deb_filelist = []
        self.controlfiles = []
        self.config_files = []
        self.params = self.rule.params

    def prep(self):
        allowed_deps = (pkgfilegroup.PkgFileGroup, pkg_symlink.PkgSymlink)
        for dep in self.rule.composed_deps():
            deprule = self.rule.subgraph.node[dep]['target_obj']
            if not isinstance(deprule, allowed_deps):
                if (isinstance(deprule, filegroup.FileGroup)
                        and deprule.address
                        in self.rule.allowed_filegroup_targets):
                    pass
                else:
                    raise error.InvalidRule(
                        'In %s: gendeb rules can only depend on pkgfilegroup '
                        'and/or pkg_symlink targets. Was given %s, '
                        'which is a %s' % (self.address, repr(dep),
                                           type(deprule).__name__))
        base.BaseBuilder.prep(self)

    def build(self):
        params = self.params
        deb_filename = os.path.basename(self.rule.output_files[0])
        maintainer = params['packager'] or '<%s@%s>' % (
            pwd.getpwuid(os.getuid()).pw_name, socket.gethostname())
        fpm_description = [params['short_description']
                           ] + params['long_description']
        # The required args:
        cmd = [self.fpm_bin, '-f',
               '-t', 'deb',
               '-s', 'dir',
               '--package', deb_filename,
               '--name', params['package_name'],
               '--version', params['version'],
               '--iteration', params['release'],
               '--deb-user', params['files_owner'],
               '--deb-group', params['files_group'],
               '--description', '\n'.join(fpm_description)]
        # Optional things that have default values:
        cmd.extend([
            '--maintainer', maintainer,
            '--epoch', params['epoch'],
            '--category', params['section'],
            '--architecture', params['arch'],
            '--deb-priority', params['priority'],
            ])
        # Optional parameters:
        if params['extra_requires']:
            cmd.extend(util.repeat_flag(params['extra_requires'], '--depends'))
        if params['conflicts']:
            cmd.extend(util.repeat_flag(params['conflicts'], '--conflicts'))
        if params['homepage']:
            cmd.extend(['--url', params['homepage']])

        maintainer_script = lambda x: (self.rulefor(params[x])
                                           .source_files.next())
        # See that next()? source_files is a generator.
        # Otherwise that would be [0].

        if params['postinst']:
            cmd.extend(['--after-install', maintainer_script('postinst')])
        if params['postrm']:
            cmd.extend(['--after-remove', maintainer_script('postrm')])
        if params['preinst']:
            cmd.extend(['--before-install', maintainer_script('preinst')])
        if params['prerm']:
            cmd.extend(['--before-remove', maintainer_script('prerm')])
        if params['extra_control_fields']:
            for field, val in params['extra_control_fields']:
                cmd.extend(['--deb-field', '%s: %s' % (field, val)])

        if self.config_files:
            cmd.extend(util.repeat_flag(self.config_files, '--config-files'))

        if self.deb_filelist:
            inputs_file = os.path.join(self.workdir, '__inputs')
            with open(inputs_file, 'w') as ifh:
                ifh.write('\n'.join(self.deb_filelist))
                ifh.write('\n')
            cmd.extend(['-C', self.deb_fsroot,
                        '--inputs', inputs_file])

        # string out those pesky integers
        cmd = [str(x) for x in cmd]

        log.debug('Generated fpm command: %s', cmd)
        ruledir = os.path.join(self.buildroot, self.address.repo,
                               self.address.path)
        fpm = subprocess.Popen(cmd, stdout=sys.stdout, stderr=sys.stderr,
                               cwd=ruledir)
        fpm.wait()
        with open(os.path.join(self.buildroot, self.rule.output_files[1]),
                  'w') as changesfile:
            changesfile.write(self.genchanges())

    #def collect_srcs(self):
    #    pass

    def genchanges(self):
        """Generate a .changes file for this package."""
        chparams = self.params.copy()
        debpath = os.path.join(self.buildroot, self.rule.output_files[0])
        chparams.update({
            'fullversion': '{epoch}:{version}-{release}'.format(**chparams),
            'metahash': self._metahash().hexdigest(),
            'deb_sha1': util.hash_file(debpath, hashlib.sha1()).hexdigest(),
            'deb_sha256': util.hash_file(debpath, hashlib.sha256()
                                         ).hexdigest(),
            'deb_md5': util.hash_file(debpath, hashlib.md5()).hexdigest(),
            'deb_bytes': os.stat(debpath).st_size,
            # TODO: having to do this split('/')[-1] is absurd:
            'deb_filename': debpath.split('/')[-1],
            })

        output = '\n'.join([
            'Format: 1.8',
            # Static date string for repeatable builds:
            'Date: Tue, 01 Jan 2013 00:00:00 -0700',
            'Source: {package_name}',
            'Binary: {package_name}',
            'Architecture: {arch}',
            'Version: {fullversion}',
            'Distribution: {distro}',
            'Urgency: {urgency}',
            'Maintainer: {packager}',
            'Description: ',
            ' {package_name} - {short_description}',
            'Changes: ',
            ' {package_name} ({fullversion}) {distro}; urgency={urgency}',
            ' .',
            ' * Built by Butcher - metahash for this build is {metahash}',
            'Checksums-Sha1: ',
            ' {deb_sha1} {deb_bytes} {deb_filename}',
            'Checksums-Sha256: ',
            ' {deb_sha256} {deb_bytes} {deb_filename}',
            'Files: ',
            ' {deb_md5} {deb_bytes} {section} {priority} {deb_filename}',
            ''  # Newline at end of file.
            ]).format(**chparams)

        return output

    def collect_deps(self):
        deb_filelist = []
        for tgt in self.rule.composed_deps() or []:
            rule = self.rulefor(tgt)
            for item in rule.output_files:
                item_base = item.lstrip('/').split(
                    os.path.join(rule.address.repo,
                                 rule.address.path), 1)[-1].lstrip('/')
                # Maintainer scripts and such don't need to be copied
                # into the tree:
                # (they'll be picked up from their location in situ)
                if rule.address not in self.rule.control_deps:
                    deb_filelist.append(item_base)
                    if ('section' in rule.params
                            and rule.params['section'] == 'config'):
                        self.config_files.append('/%s' % item_base)
                    dst = self.deb_fsroot
                    self.linkorcopy(
                        os.path.join(self.buildroot, item),
                        os.path.join(dst, item_base))
        self.deb_filelist = deb_filelist


class GenDeb(base.BaseTarget):
    """gendeb rule"""

    rulebuilder = GenDebBuilder
    ruletype = 'gendeb'

    required_params = [
        ('name', str),
        ('version', (str, int)),
        ('release', (str, int)),
        ('long_description', list),
        ('short_description', str),
        ]
    optional_params = [
        ('arch', str, 'amd64'),
        ('conflicts', list, None),
        ('deps', list, None),
        ('distro', str, 'unstable'),
        ('epoch', (str, int), 0),
        ('extra_control_fields', list, None),
        ('extra_requires', list, None),
        ('files_owner', str, 'root'),
        ('files_group', str, 'root'),
        ('homepage', str, None),
        ('packager', str, None),
        ('package_name', str, None),
        ('priority', str, 'extra'),
        ('postinst', str, None),
        ('postrm', str, None),
        ('preinst', str, None),
        ('prerm', str, None),
        ('section', str, 'misc'),
        ('urgency', str, 'low'),
        # Not ready for these:
        #('strip', bool, False),
        #('triggers', str, None),
        ]

    def __init__(self, **kwargs):
        self.params = {}
        base.BaseTarget.__init__(self, **kwargs)
        if not self.params['package_name']:
            self.params['package_name'] = self.params['name']
        # Exceptions to the "only pkgfilegroup or pkg_symlink" restriction so
        # preinst/postinst/prerm/postrm dependencies work:
        self.allowed_filegroup_targets = set()
        self.control_deps = set()

    def validate_args(self):
        """Input validators for this rule type."""
        base.BaseTarget.validate_args(self)
        params = self.params
        if params['extra_control_fields'] is not None:
            assert isinstance(params['extra_control_fields'], list), (
                'extra_control_fields must be a list of tuples, not %s' % type(
                    params['extra_control_fields']))
            for elem in params['extra_control_fields']:
                assert (isinstance(elem, tuple) and len(elem) == 1), (
                    'extra_control_fields must be a list of 2-element tuples. '
                    'Invalid contents: %s' % elem)
        pkgname_re = '^[a-z][a-z0-9+-.]+'
        assert re.match(pkgname_re, params['package_name']), (
            'Invalid package name: %s. Must match %s' % (
                params['package_name'], pkgname_re))
        # TODO: more of this.

    def composed_deps(self):
        deps = []
        for dep in ['preinst', 'postinst', 'prerm', 'postrm']:
            if self.params[dep]:
                dep_addr = self.makeaddress(self.params[dep])
                deps.append(dep_addr)
                self.control_deps.add(dep_addr)
                self.allowed_filegroup_targets.add(dep_addr)
        return deps + base.BaseTarget.composed_deps(self)

    @property
    def output_files(self):
        return [
            os.path.join(
                self.address.repo, self.address.path,
                '{package_name}_{version}-{release}_{arch}.deb'.format(
                    **self.params)),
            os.path.join(
                self.address.repo, self.address.path,
                '{package_name}_{version}-{release}_{arch}.changes'.format(
                    **self.params))
            ]



"""Unimplemented target class."""

from butcher.targets import base
from twitter.common import log


class UnimplementedTarget(base.BaseTarget):
    ruletype = 'UNKNOWN'

    def __init__(self, name, *args, **kwargs):
        log.warn('New Unimplemented %s target: name=%s, %s, %s',
                 self.ruletype, name, args, kwargs)
        base.BaseTarget.__init__(self, name)



#!/usr/bin/env python2.7
# Copyright 2013 Cloudscaling Inc.
#
# This is experimental and incomplete. Don't judge me :-P

"""Butcher: a distributed build system."""

__author__ = 'Benjamin Staffin <benley@gmail.com>'

# If you want this, it has to happen before importing gitrepo:
#os.environ.update({'GIT_PYTHON_TRACE': 'full'})

import networkx
import os
import pprint
import shutil
from twitter.common import log
from twitter.common import app
from butcher import buildfile
from butcher import address
from butcher import cache
from butcher import error
from butcher import gitrepo
from butcher import util
from butcher.targets import base

app.add_option('--debug', action='store_true', dest='debug')
app.add_option(
    '--basedir',
    dest='butcher_basedir',
    help='Base directory for butcher to work in.',
    default=os.path.join(util.user_homedir(), '.cache', 'butcher'))
app.add_option(
    '--build_root',
    dest='build_root',
    help=('Base directory in which builds will be done. If unspecified, makes '
          'a build directory inside of the butcher basedir.'))
app.add_option(
    '--buildfile_name',
    dest='buildfile_name',
    help='Filename to use as BUILD files in each directory.',
    default='BUILD')
app.add_option(
    '--rebuild_all',
    action='store_true',
    dest='disable_cache_fetch',
    help='Disable cache fetching and explicitly build each target.')
app.add_option(
    '--nohardlinks',
    action='store_true',
    dest='disable_hardlinks',
    help='Disable hardlinking of files in and out of the cache.')
app.add_option(
    '--outdir',
    dest='final_output_dir',
    help='Copy the output file(s) of the final target to this directory.')


class Butcher(app.Module):
    """Butcher!"""

    options = {'cache_fetch': True}

    def __init__(self):
        app.Module.__init__(self, label='butcher',
                            description='Butcher build system.',
                            dependencies='twitter.common.log')
        self.repo_state = gitrepo.RepoState()
        self.graph = networkx.DiGraph()
        # TODO: there is no good reason to keep all the separate subgraphs.
        self.subgraphs = {}
        self.failure_log = []  # Build failure exceptions get kept in here.
        self.buildroot = None

    def setup_function(self):
        """Runs prior to the global main function."""
        log.options.LogOptions.set_stderr_log_level('google:INFO')
        if app.get_options().debug:
            log.options.LogOptions.set_stderr_log_level('google:DEBUG')
        if not app.get_options().build_root:
            app.set_option('build_root', os.path.join(
                app.get_options().butcher_basedir, 'build'))
        self.buildroot = app.get_options().build_root
        if not os.path.exists(self.buildroot):
            os.makedirs(self.buildroot)
        if app.get_options().disable_cache_fetch:
            self.options['cache_fetch'] = False
        if app.get_options().disable_hardlinks:
            base.BaseBuilder.linkfiles = False

    def clean(self):
        """Clear the contents of the build area."""
        if os.path.exists(self.buildroot):
            log.info('Clearing the build area.')
            log.debug('Deleting: %s', self.buildroot)
            shutil.rmtree(self.buildroot)
            os.makedirs(self.buildroot)

    def build(self, explicit_target):
        if explicit_target not in self.graph.nodes():
            raise error.NoSuchTargetError('No rule defined for %s' %
                                          explicit_target)
        if not self.options['cache_fetch']:
            log.info('Cache fetching disabled. Doing a clean build.')

        # Get the subgraph of only the things we need built.
        # (yes, topological sort accomplishes that)
        buildgraph = self.graph.subgraph(
            networkx.topological_sort(self.graph, nbunch=[explicit_target]))
        #if app.get_options().debug:
        #    log.debug('Buildgraph edges:\n%s',
        #              pprint.pformat(buildgraph.edges()))
        #    log.debug('Buildgraph nodes:\n%s',
        #              pprint.pformat(buildgraph.node))

        # TODO: this should be parallelized.

        # Caching notes:
        # - topological sort:
        #   - (that is, start at the root of the tree)
        #   - For each node,
        #     - IF nothing depends on it (i.e. it has no predecessors)
        #       AND it is not the explicit build target,
        #       - Remove it from the build tree (which may create more orphans
        #         to deal with in a later iteration)
        #     - ELSEIF it is in the build cache,
        #       - Mark it as built
        #       - Remove it from the build tree (which may create orphans to
        #         deal with in a later iteration)
        #     - ELSE:
        #       - Build it!
        #       - Send its outputs to the build cache
        #       - If it was the explicit build target, exit the loop.
        #       - Otherwise, remove it from the build tree and continue
        # - Next, if the explicit target hasn't already been built
        #   at this point,
        #   - Do a fresh *reverse* topological sort of the build graph,
        #     so nodes get built before the things that depend upon them.
        #   - Iterate over that list:
        #     - Build each node
        #     - Send it to the build cache
        #     - Remove it from the build tree
        # ... and at the end of all that the requested thing has been built.

        # This is probably going to end up iterating over all the nodes twice,
        # followed by a recursive build process that could technically
        # accomplish the same thing, but this method will reduce recursion
        # depth. Not sure if that will ever actually matter.

        # Iterate from the top of the tree, pruning based on
        # cached/built status.
        #buildlist = networkx.topological_sort(buildgraph)
        #for node in buildlist:
        #  if node != explicit_target and not buildgraph.predecessors(node):
        #    # It's an orphaned node and we don't need it.
        #    buildgraph.remove_node(node)
        #    continue
        #  if self.already_built(node):
        #    # It's already built (or cached)
        #    buildgraph.remove_node(node)
        #    if node == explicit_target:
        #      # The explicitly requested target has already been built.
        #      # Groovy.
        #      node_rule.get_from_cache()
        #      buildlist = []
        #      break

        # Now that we've pruned the tree, start building from the _bottom_.
        buildlist = networkx.topological_sort(buildgraph)
        buildlist.reverse()
        if buildlist:  # but sure there's actually anything left to do first.
            for node in buildlist:
                try:
                    node_obj = buildgraph.node[node]['target_obj']
                    node_builder = node_obj.rulebuilder(
                        self.buildroot, node_obj,
                        # TODO: this is absurd:
                        self.repo_state.GetRepo(node.repo).repo.tree().abspath)
                    node_builder.prep()
                    if self.options['cache_fetch']:
                        try:
                            node_builder.get_from_cache()
                            log.info('[%s]: cache hit', node)
                        except cache.CacheMiss as err:
                            log.debug('[%s]: cache miss: %s', node, err)
                            log.info('[%s]: Building.', node)
                            node_builder.build()
                            node_builder.collect_outs()
                    else:
                        log.info('[%s]: Building.', node)
                        node_builder.build()
                        node_builder.collect_outs()
                except (error.BuildFailure, error.InvalidRule) as err:
                    log.error('[%s]: failed: %s', node, err)
                    self.failure_log.append(err)
                    break
                else:
                    log.debug('[%s]: Build succeeded.', node)
                    buildgraph.remove_node(node)

        if buildgraph.nodes():  # If the list isn't empty, the build failed.
            raise error.OverallBuildFailure(
                'Build failed due to previous errors.')
        else:
            log.info('Success.', explicit_target)
            outputs = list(
                self.graph.node[explicit_target]['target_obj'].output_files)
            if not outputs:
                return

            # The following section is horrible code. Sorry.
            desired_outdir = app.get_options().final_output_dir
            finalfiles = []

            if desired_outdir:
                bases = set(os.path.dirname(f) for f in outputs)
                if len(bases) > 1:
                    strip_prefix = None
                    log.warn('Output files are in multiple directories. '
                             '*NOT* flattening.')
                else:
                    strip_prefix = bases.pop()
                    if not strip_prefix.endswith('/'):
                        strip_prefix = strip_prefix + '/'

                for item in outputs:
                    builtfile = os.path.join(self.buildroot, item)
                    if strip_prefix:
                        if item.startswith(strip_prefix):
                            item = item[len(strip_prefix):]
                    finalfile = os.path.join(desired_outdir, item)
                    util.linkorcopy(builtfile, finalfile)
                    finalfiles.append(finalfile)
            else:  # No desired_outdir set.
                for item in outputs:
                    finalfiles.append(os.path.join(self.buildroot, item))

            if finalfiles:
                log.info('Outputs:')
                for item in finalfiles:
                    log.info('  %s', item)

    def load_graph(self, startingpoint):
        s_tgt = address.new(startingpoint, target='all')
        log.info('Loading graph starting at %s', s_tgt)
        s_subgraph = buildfile.load(self.load_buildfile(s_tgt),
                                    s_tgt.repo, s_tgt.path)
        self.subgraphs[s_tgt] = s_subgraph
        self.graph = networkx.compose(self.graph, s_subgraph)

        while self.paths_wanted:
            log.debug('Loaded so far: %s', self.paths_loaded)
            log.debug('Unresolved nodes: %s', self.missing_nodes)
            n_tgt = self.paths_wanted.pop()
            if n_tgt in self.paths_loaded:
                mlist = ', '.join(str(x) for x in self.missing_nodes)
                raise error.BrokenGraph('Broken graph! Missing targets: %s' %
                                        mlist)
            n_subgraph = buildfile.load(self.load_buildfile(n_tgt),
                                        n_tgt.repo, n_tgt.path)
            self.subgraphs[n_tgt] = n_subgraph
            # Replace "missing" nodes with actual nodes:
            for node in self.missing_nodes.intersection(n_subgraph.nodes()):
                self.graph.node[node].update(n_subgraph.node[node])
            # Add the new nodes
            # (attributes in self.graph take precedence here):
            self.graph = networkx.compose(self.graph, n_subgraph)

        # Traverse the graph and attach subgraphs to each node
        for node in self.graph.node:
            node_attrs = self.graph.node[node]
            subgraph = self.graph.subgraph(
                networkx.topological_sort(self.graph, nbunch=[node]))
            node_attrs['target_obj'].subgraph = subgraph

    @property
    def paths_loaded(self):
        """List of paths already visited and loaded."""
        return self.subgraphs.keys()

    @property
    def paths_wanted(self):
        """The set of paths where we expect to find missing nodes."""
        return set(address.new(b, target='all') for b in self.missing_nodes)

    @property
    def missing_nodes(self):
        """The set of targets known as dependencies but not yet defined."""
        missing = set()
        for target_addr, target_attrs in self.graph.node.items():
            if 'target_obj' not in target_attrs:
                missing.add(target_addr)
        return missing

    def load_buildfile(self, target):
        """Pull a build file from git."""
        log.info('Loading: %s', target)
        filepath = os.path.join(target.path, app.get_options().buildfile_name)
        try:
            repo = self.repo_state.GetRepo(target.repo)
            return repo.get_file(filepath)
        except gitrepo.GitError as err:
            log.error('Failed loading %s: %s', target, err)
            raise error.BrokenGraph('Sadface.')

    def already_built(self, target):
        """Stub. Always returns False."""
        # FIXME: implement or obviate.
        # This may end up in a cache server interface class.
        _ = target.repo
        return False


@app.command
def resolve(args):
    """Just print the result of parsing a target string."""
    if not args:
        log.error('Exactly 1 argument is required.')
        app.quit(1)
    print(address.new(args[0]))


@app.command
def build(args):
    """Build a target and its dependencies."""

    if len(args) != 1:
        log.error('One target required.')
        app.quit(1)

    target = address.new(args[0])
    log.info('Resolved target to: %s', target)

    try:
        bb = Butcher()
        bb.clean()
        bb.load_graph(target)
        bb.build(target)
    except (gitrepo.GitError,
            error.BrokenGraph,
            error.NoSuchTargetError) as err:
        log.fatal(err)
        app.quit(1)
    except error.OverallBuildFailure as err:
        log.fatal(err)
        log.fatal('Error list:')
        [log.fatal('  [%s]: %s', e.node, e) for e in bb.failure_log]
        app.quit(1)


@app.command
def rebuild(args):
    """Rebuild a target and deps, even if it has been built and cached."""
    if len(args) != 1:
        log.fatal('One target required.')
        app.quit(1)

    app.set_option('disable_cache_fetch', True)
    Butcher.options['cache_fetch'] = False
    build(args)


@app.command
def clean(args):
    """Akin to make clean"""
    bb = Butcher()
    bb.clean()


@app.command
def dump(args):
    """Load the build graph for a target and dump it to stdout."""
    if len(args) != 1:
        log.error('One target required.')
        app.quit(1)

    try:
        bb = Butcher()
        bb.load_graph(args[0])
    except error.BrokenGraph as lolno:
        log.fatal(lolno)
        app.quit(1)
    print "Nodes:"
    pprint.pprint(bb.graph.node)
    print "Edges:"
    pprint.pprint(bb.graph.edge)


@app.command
def draw(args):
    """Load the build graph for a target and render it to an image."""
    if len(args) != 2:
        log.error('Two arguments required: [build target] [output file]')
        app.quit(1)

    target = args[0]
    out = args[1]

    try:
        bb = Butcher()
        bb.load_graph(target)
    except error.BrokenGraph as lolno:
        log.fatal(lolno)
        app.quit(1)

    # Filter down to the target and all of its transitive dependencies.
    # TODO: make it possible to optionally draw the entire graph
    filtered_graph = bb.graph.subgraph(
        networkx.topological_sort(bb.graph, nbunch=[address.new(target)]))

    a = networkx.to_agraph(filtered_graph)
    a.draw(out, prog='dot')
    log.info('Graph written to %s', out)


def stub_main():
    """setuptools blah: it still can't run a module as a script entry_point"""
    from google.apputils import run_script_module
    import butcher.main
    run_script_module.RunScriptModule(butcher.main)

if __name__ == '__main__':
    app.register_module(Butcher())
    app.register_module(gitrepo.RepoState())
    app.register_module(cache.CacheManager())
    app.interspersed_args(True)
    app.main()



"""Build rule representation."""

import re
from butcher import error

# These are manipulated by PythonBuildFile to set context.
CUR_REPO = None
CUR_PATH = ''


def new(*args, **kwargs):
    """Return a new Address object."""
    return Address(*args, **kwargs)


class Address(dict):
    """A build rule address.

    Mostly acts like a dictionary, except when you need it to act
    like a string.
    """

    # The components that make up a build target:
    params = ('repo', 'git_ref', 'path', 'target')

    def __init__(self, *args, **kwargs):
        dict.__init__(self, {'repo': None,
                             'git_ref': '',
                             'path': '',
                             'target': 'all'})
        self.update(*args, **kwargs)
        if not self.repo and CUR_REPO is not None:
            self.repo = str(CUR_REPO)  # str() to ensure copy, not reference
        if not self.path:
            self.path = str(CUR_PATH)

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def encode(self, *args, **kwargs):
        """str.encode"""
        return str(self).encode(*args, **kwargs)
    encode.__doc__ = str.encode.__doc__

    def update(self, iterable=None, **kwargs):
        if iterable:
            try:
                for k in iterable.keys():
                    self[k] = iterable[k]
            except AttributeError:
                try:
                    for (k, val) in iterable:
                        self[k] = val
                except (TypeError, ValueError):
                    self.update(self.__parse_target(iterable))
        for k in kwargs:
            self[k] = kwargs[k]

    def __setitem__(self, key, value):
        if key not in ('repo', 'git_ref', 'path', 'target'):
            raise error.ButcherError('Invalid parameter: %s' % key)
        # TODO: value validation
        dict.__setitem__(self, key, value)

    def __repr__(self):
        if self.repo is None or self.target is None:
            raise error.AddressError(
                'No valid repr for this incomplete target: %s' % dict(self))

        ret = ['//%(repo)s' % dict(self)]
        if self['path']:
            ret.append('/%s' % self['path'])
        ret.append(':%(target)s' % self)
        return ''.join(ret)

    @property
    def repo(self):
        return self['repo']

    @repo.setter
    def repo(self, val):
        self['repo'] = val

    @property
    def git_ref(self):
        return self['git_ref']

    @git_ref.setter
    def git_ref(self, val):
        self['git_ref'] = val

    @property
    def path(self):
        return self['path']

    @path.setter
    def path(self, val):
        self['path'] = val

    @property
    def target(self):
        return self['target']

    @target.setter
    def target(self, val):
        self['target'] = val

    @staticmethod
    def __parse_target(targetstr, current_repo=None):
        """Parse a build target string.

        General form: //repo[gitref]/dir/path:target.

        These are all valid:
          //repo
          //repo[a038fi31d9e8bc11582ef1b1b1982d8fc]
          //repo[a039aa30853298]:foo
          //repo/dir
          //repo[a037928734]/dir
          //repo/dir/path
          //repo/dir/path:foo
          :foo
          dir/path
          dir/path:foo
          dir:foo

        Returns: {'repo': '//reponame',
                  'git_ref': 'a839a38fd...',
                  'path': 'dir/path',
                  'target': 'targetname}
        """
        # 'blah' -> ':blah'
        if not (':' in targetstr or '/' in targetstr):
            targetstr = ':%s' % targetstr

        match = re.match(
            r'^(?://(?P<repo>[\w-]+)(?:\[(?P<git_ref>.*)\])?)?'
            r'(?:$|/?(?P<path>[\w/-]+)?(?::?(?P<target>[\w-]+)?))', targetstr)
        try:
            groups = match.groupdict()
            if not groups['repo']:
                groups['repo'] = current_repo
            if not groups['git_ref']:
                groups['git_ref'] = 'develop'
            if not groups['target']:
                groups['target'] = 'all'
            if not groups['path']:
                groups['path'] = ''
        except AttributeError:
            raise error.ButcherError('"%s" is not a valid build target.')
        #log.debug('parse_target: %s -> %s', targetstr, groups)
        return groups



"""Parsing context manager for BUILD files.

Things available in BUILD context:
  invokable rules:
    - At the time of writing this: genrule, gendeb, filegroup, pkgfilegroup
    - See documentation for each type.
    - For a list see butcher.targets.__init__.py
  functions:
    - glob: see python help for glob.glob
  variables:
    - ROOTDIR: path to the source tree root.
    - RULEDIR: path to the directory the BUILD file is in.
"""

import collections
import contextlib
import copy
import os
from butcher import error


class ContextError(error.ButcherError):
    """Something requiring a BUILD context was attempted out of context."""
    pass


class ParseContext(object):
    """Manage context for parsing/loading BUILD files."""
    _active = collections.deque([])
    _parsed = set()
    # These strings get executed inside each buildfile context
    # to initialize it:
    _strs_to_exec = [
        'from butcher.target_context import *',
        ]

    @staticmethod
    @contextlib.contextmanager
    def activate(ctx):
        """Activate the given ParseContext."""
        if hasattr(ctx, '_on_context_exit'):
            raise ContextError(
                'Context actions registered outside this '
                'parse context are active')

        try:
            ParseContext._active.append(ctx)
            ctx._on_context_exit = []
            yield
        finally:
            for func, args, kwargs in ctx._on_context_exit:
                func(*args, **kwargs)
            del ctx._on_context_exit
            ParseContext._active.pop()

    def __init__(self, build_file):
        self.build_file = build_file
        self._parsed = False

    def parse(self, **global_args):
        """Entry point to parsing a BUILD file.

        Args:
          **global_args: Variables to include in the parsing environment.
        """

        if self.build_file not in ParseContext._parsed:
            # http://en.wikipedia.org/wiki/Abstract_syntax_tree
            # http://martinfowler.com/books/dsl.html
            butcher_context = {}
            for str_to_exec in self._strs_to_exec:
                ast = compile(str_to_exec, '<string>', 'exec')
                exec_function(ast, butcher_context)

            with ParseContext.activate(self):
                startdir = os.path.abspath(os.curdir)
                try:
                    os.chdir(self.build_file.path_on_disk)
                    if self.build_file not in ParseContext._parsed:
                        ParseContext._parsed.add(self.build_file)
                        eval_globals = copy.copy(butcher_context)
                        eval_globals.update(
                            {'ROOT_DIR': self.build_file.path_on_disk,
                             '__file__': 'bogus please fix this'})
                        eval_globals.update(global_args)
                        exec_function(self.build_file.code, eval_globals)
                finally:
                    os.chdir(startdir)


def exec_function(ast, globals_map):
    """Execute a python code object in the given environment.

    Args:
      globals_map: Dictionary to use as the globals context.
    Returns:
      locals_map: Dictionary of locals from the environment after execution.
    """
    locals_map = globals_map
    exec ast in globals_map, locals_map
    return locals_map



"""pkg_symlink targets."""

from butcher.targets import base


class PkgSymlinkBuilder(base.BaseBuilder):
    pass


class PkgSymlink(base.BaseTarget):
    rulebuilder = PkgSymlinkBuilder
    ruletype = 'pkg_symlink'

    required_params = [
        ('name', str),
        ('link_name', str),
        ('link_target', str)]
    optional_params = [
        ('mode', int, None),
        ('owner', str, None),
        ('group', str, None)]



"""Virtual (deps only) targets"""

from butcher.targets import base
from butcher import error
from twitter.common import log


class VirtualTargetBuilder(base.BaseBuilder):
    """Builder for virtual (deps only) rules."""

    def collect_srcs(self):
        pass

    def build(self):
        log.debug('[%s] nothing to build.', self.rule.address)


class VirtualTarget(base.BaseTarget):
    """Virtual (deps only) rule."""

    rulebuilder = VirtualTargetBuilder
    ruletype = 'virtual'

    required_params = [('name', str), ('deps', list)]
    optional_params = []

    def __init__(self, **kwargs):
        base.BaseTarget.__init__(self, **kwargs)
        if not kwargs['deps']:
            raise error.InvalidRule(
                'Virtual rules with no deps make no sense.')
        log.debug('New virtual target: %s, deps: %s',
                  kwargs['name'], kwargs['deps'])

    @property
    def output_files(self):
        """Returns all output files from all of the current module's rules."""
        for dep in self.subgraph.successors(self.address):
            dep_rule = self.subgraph.node[dep]['target_obj']
            for out_file in dep_rule.output_files:
                yield out_file



#!/usr/bin/python2.7
"""Print the value of a assigned var from a .py file *without* executing it."""

import ast
import os.path
import sys


def getvar(syntree, targetvar):
    """Scan an ast object for targetvar and return its value.

    Only handles single direct assignment of python literal types. See docs on
    ast.literal_eval for more info:
    http://docs.python.org/2/library/ast.html#ast.literal_eval

    Args:
      syntree: ast.Module object
      targetvar: name of global variable to return
    Returns:
      Value of targetvar if found in syntree, or None if not found.
    """
    for node in syntree.body:
        if isinstance(node, ast.Assign):
            for var in node.targets:
                if var.id == targetvar:
                    return ast.literal_eval(node.value)


def main(argv):
    if len(argv) != 3:
        print("USAGE: {} <filename.py> <variable_name>".format(
            os.path.basename(argv[0])))
        sys.exit(1)

    srcfile = argv[1]
    targetvar = argv[2]

    with open(srcfile, 'r') as src:
        code = ast.parse(src.read(), srcfile)
        print getvar(code, targetvar)


if __name__ == '__main__':
    main(sys.argv)



# Copyright 2013 Cloudscaling Inc.

"""Git repository wrapper module."""

# If you want this, it has to happen before importing git, as silly as that is:
#os.environ.update({'GIT_PYTHON_TRACE': 'full'})

import git
import gitdb
import os
from twitter.common import app
from twitter.common import log
from butcher import address

app.add_option('--repo_baseurl', dest='repo_baseurl',
               help='Base URL to git repo collection.',
               default='https://github.com/benley')
app.add_option(
    '--pin', action='append', dest='pinned_repos',
    help='Pin a repo to a particular symbolic ref. Syntax: //reponame[ref]')
app.add_option(
    '--map_repo', action='append', dest='repo_overrides',
    help=('Override the upstream location of a repo. '
          'Format: <reponame>:</path/to/repo>'))
app.add_option(
    '--default_ref', dest='default_ref', default='HEAD',
    help='Default git symbolic ref to checkout if not pinned otherwise.')

# TODO: validate repo_overrides before starting any real work.
# TODO: a way to override the working copy location?
#        (in addition to map_repo, which just changes the upstream url)


class GitError(RuntimeError):
    """Generic error class."""
    pass


class RepoState(app.Module):
    """Holds git repo state. Singleton thanks to app.Module."""
    repos = {}
    pins = {}
    origin_overrides = {}
    repo_basedir = ''

    def __init__(self):
        app.Module.__init__(self, label=__name__,
                            description='Git repo subsystem.')

    def setup_function(self):
        overrides = app.get_options().repo_overrides
        if overrides:
            for line in overrides:
                (reponame, path) = line.split(':')
                self.origin_overrides[reponame] = os.path.abspath(
                    os.path.expanduser(path))
        pins = app.get_options().pinned_repos
        for pin in (pins or []):
            ppin = address.new(pin)
            self.pins[ppin.repo] = ppin.git_ref
        self.repo_basedir = os.path.join(app.get_options().butcher_basedir,
                                         'gitrepo')

    def GetRepo(self, reponame):
        if reponame not in self.repos:
            origin = None
            if reponame in self.origin_overrides:
                origin = self.origin_overrides[reponame]
            ref = None
            if reponame in self.pins:
                ref = self.pins[reponame]
            self.repos[reponame] = GitRepo(reponame, ref, origin=origin)
        return self.repos[reponame]

    def HeadList(self):
        """Return a list of all the currently loaded repo HEAD objects."""
        return [(rname, repo.currenthead) for rname, repo in self.repos.items()
                ]


class GitRepo(object):
    """Git repository wrapper."""

    def __init__(self, name, ref=None, origin=None):
        opts = app.get_options()
        self.repo_baseurl = opts.repo_baseurl
        #log.debug('Base url: %s', self.repo_baseurl)

        self.repo_basedir = RepoState().repo_basedir
        #log.debug('Base directory: %s', self.repo_basedir)

        self.name = name
        #log.debug('Repo name: %s', self.name)

        self.ref = ref or opts.default_ref

        if origin:
            self.origin_url = origin
        else:
            self.origin_url = '%s/%s' % (self.repo_baseurl, self.name)
        log.debug('[%s] Origin url: %s', self.name, self.origin_url)

        self.repodir = os.path.join(self.repo_basedir, self.name)
        log.debug('[%s] Working copy: %s', self.name, self.repodir)

        try:
            self.repo = git.Repo(self.repodir)
        except git.exc.NoSuchPathError:
            self.repo = git.Repo.init(self.repodir, bare=False)
            log.debug('[%s] Repo initialized.', self.name)
        except git.exc.InvalidGitRepositoryError:
            log.error(
                "[%s] %s exists, but is not a valid git repo. Can't continue.",
                self.name, self.repodir)
            raise GitError

        self.setorigin()
        self.fetchref(self.ref)
        self.sethead(self.ref)

    def setorigin(self):
        """Set the 'origin' remote to the upstream url that we trust."""
        try:
            origin = self.repo.remotes.origin
            if origin.url != self.origin_url:
                log.debug('[%s] Changing origin url. Old: %s New: %s',
                          self.name, origin.url, self.origin_url)
                origin.config_writer.set('url', self.origin_url)
        except AttributeError:
            origin = self.repo.create_remote('origin', self.origin_url)
            log.debug('[%s] Created remote "origin" with URL: %s',
                      self.name, origin.url)

    def fetchall(self):
        """Fetch all refs from the upstream repo."""
        try:
            self.repo.remotes.origin.fetch()
        except git.exc.GitCommandError as err:
            raise GitError(err)

    def fetchref(self, ref):
        """Fetch a particular git ref."""
        log.debug('[%s] Fetching ref: %s', self.name, ref)
        fetch_info = self.repo.remotes.origin.fetch(ref).pop()
        return fetch_info.ref

    def sethead(self, ref):
        """Set head to a git ref."""
        log.debug('[%s] Setting to ref %s', self.name, ref)
        try:
            ref = self.repo.rev_parse(ref)
        except gitdb.exc.BadObject:
            # Probably means we don't have it cached yet.
            # So maybe we can fetch it.
            ref = self.fetchref(ref)
        log.debug('[%s] Setting head to %s', self.name, ref)
        self.repo.head.reset(ref, working_tree=True)
        log.debug('[%s] Head object: %s', self.name, self.currenthead)

    @property
    def currenthead(self):
        """Returns the current head object."""
        return self.repo.head.object

    def get_file(self, filename):
        """Get a file from the repo.

        Returns a file-like stream with the data.
        """
        log.debug('[%s]: reading: //%s/%s', self.name, self.name, filename)
        try:
            blob = self.repo.head.commit.tree/filename
            return blob.data_stream
        except KeyError as err:
            raise GitError(err)



"""Importable context to be used inside BUILD files."""

# Things that don't go into the build context:
from twitter.common import log

# Breaking my own rule of only importing whole modules in order to set up a
# sublanguage parser. This is an unusual case.

# Butcher targets:
from butcher.targets.gendeb import GenDeb as gendeb
from butcher.targets.genrule import GenRule as genrule
from butcher.targets.filegroup import FileGroup as filegroup
from butcher.targets.pkgfilegroup import PkgFileGroup as pkgfilegroup
from butcher.targets.pkg_symlink import PkgSymlink as pkg_symlink
#from butcher.targets.virtual import VirtualTarget as virtual

# Other useful things to have in there:
from butcher.util import glob

__all__ = ['gendeb', 'genrule', 'filegroup', 'pkgfilegroup', 'pkg_symlink',
           'glob', 'globs', 'rglobs']


def globs(*args):
    """Deprecated: use glob() instead."""
    # TODO: use the warnings module to issue deprecation warnings.
    log.warn('globs() is deprecated; please use glob() instead.')
    return glob(*args)


def rglobs(*args):
    """Deprecated: use glob() instead."""
    log.warn('rglobs() is deprecated; please use glob() instead.')
    return glob(*args)



"""Miscellaneous utility functions."""

import collections
import glob2
import hashlib
import os
import shutil
from butcher import error
from twitter.common import log


def user_homedir(username=None):
    """Returns a user's home directory.

    If no username is specified, returns the current user's homedir.
    """
    if username:
        return os.path.expanduser('~%s/' % username)
    elif 'HOME' in os.environ:
        return os.environ['HOME']
    elif os.name == 'posix':
        return os.path.expanduser('~/')
    else:
        raise RuntimeError('This function has failed at life.')


def hash_file(filename, hasher=None, blocksize=65536):
    """Checksum a file, optionally updating an existing hash.

    Args:
      filename: Path to the file.
      hasher: Hash object such as hashlib.sha1(). Defaults to sha1.
      blocksize: Read from the file this many bytes at a time.
    """
    return hash_stream(open(filename, 'rb'), hasher, blocksize)


def hash_stream(fileobj, hasher=None, blocksize=65536):
    """Read from fileobj stream, return hash of its contents.

    Args:
      fileobj: File-like object with read()
      hasher: Hash object such as hashlib.sha1(). Defaults to sha1.
      blocksize: Read from fileobj this many bytes at a time.
    """
    hasher = hasher or hashlib.sha1()
    buf = fileobj.read(blocksize)
    while buf:
        hasher.update(buf)
        buf = fileobj.read(blocksize)
    return hasher


def hash_str(data, hasher=None):
    """Checksum hash a string."""
    hasher = hasher or hashlib.sha1()
    hasher.update(data)
    return hasher


def repeat_flag(seq, flag):
    """Intersperse flag as a predecessor to each element of seq.

    Example:
    seq = ['foo', 'bar', 'bas']
    flag = '--word'
    print repeat_flag(seq, flag)
      ==>  ['--word', 'foo', '--word', 'bar', '--word', 'bas']
    """
    for item in iter(seq):
        yield flag
        yield item


def glob(*args):
    """Returns list of paths matching one or more wildcard patterns.

    Args:
      include_dirs: Include directories in the output
    """
    if len(args) is 1 and isinstance(args[0], list):
        args = args[0]
    matches = []
    for pattern in args:
        for item in glob2.glob(pattern):
            if not os.path.isdir(item):
                matches.append(item)
    return matches


def flatten(listish):
    """Flatten an arbitrarily-nested list of strings and lists.

    Works for any subclass of basestring and any type of iterable.
    """
    for elem in listish:
        if (isinstance(elem, collections.Iterable)
                and not isinstance(elem, basestring)):
            for subelem in flatten(elem):
                yield subelem
        else:
            yield elem


def linkorcopy(src, dst):
    """Hardlink src file to dst if possible, otherwise copy."""
    if not os.path.isfile(src):
        raise error.ButcherError('linkorcopy called with non-file source. '
                                 '(src: %s  dst: %s)' % src, dst)
    elif os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    elif os.path.exists(dst):
        os.unlink(dst)
    elif not os.path.exists(os.path.dirname(dst)):
        os.makedirs(os.path.dirname(dst))

    try:
        os.link(src, dst)
        log.debug('Hardlinked: %s -> %s', src, dst)
    except OSError:
        shutil.copy2(src, dst)
        log.debug('Couldn\'t hardlink. Copied: %s -> %s', src, dst)



"""Unit tests for utility functions."""

import unittest

from butcher import util


class TestUtils(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testFlatten(self):
        """Verify that flatten() handles variously-nested iterables."""

        self.assertListEqual(
            list(util.flatten(["one", "two", ["three", "four"], "five"])),
            ["one", "two", "three", "four", "five"])

        self.assertListEqual(
            list(util.flatten(["a", "b", "c", "d", "e"])),
            ["a", "b", "c", "d", "e"])

        self.assertListEqual(
            list(util.flatten((["a", "b"], ["c", "d", ("e", "f"), "g"]))),
            ["a", "b", "c", "d", "e", "f", "g"])

        self.assertListEqual(
            list(util.flatten(["a", ["b", ["c", ["d", ["e", ["f", ["g"]]]]]]])
                 ),
            ["a", "b", "c", "d", "e", "f", "g"])


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestUtils)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""pkgfilegroup targets

Mostly implemented. Remaining functionality may be left to consumers of this
rule.

Works so far:
  - collects output files from rules given in srcs, emits them as its own
    outputs
  - Applies strip_prefix, puts files in the prefixed directory structure.

Not yet implemented:
  - set attributes
"""

import os
from butcher import address
from butcher.targets import base


class PkgFileGroupBuilder(base.BaseBuilder):
    """Builder for pkgfilegroup"""
    ruletype = 'pkgfilegroup'

    def build(self):
        for dep in self.rule.composed_deps():
            dep_rule = self.rule.subgraph.node[dep]['target_obj']
            for dep_file in dep_rule.output_files:
                src = os.path.join(self.buildroot, dep_file)
                dst = self.rule.translate_path(dep_file, dep_rule)
                output_dst = os.path.join(self.buildroot, dst)
                output_dstdir = os.path.dirname(output_dst)
                if not os.path.exists(output_dstdir):
                    os.makedirs(output_dstdir)
                self.linkorcopy(src, output_dst)
                #TODO: attrs


class PkgFileGroup(base.BaseTarget):
    """pkgfilegroup rule"""
    rulebuilder = PkgFileGroupBuilder
    ruletype = 'pkgfilegroup'

    required_params = [('name', str),
                       ('prefix', str),
                       ('srcs', list)]
    optional_params = [
        ('attr', list, None),
        ('section', str, None),  # one of ('', 'doc', 'config')
        ('strip_prefix', str,  None)]

    def translate_path(self, dep_file, dep_rule):
        """Translate dep_file from dep_rule into this rule's output path."""
        dst_base = dep_file.split(os.path.join(dep_rule.address.repo,
                                               dep_rule.address.path), 1)[-1]
        if self.params['strip_prefix']:
            dst_base = dep_file.split(self.params['strip_prefix'], 1)[-1]
        return os.path.join(self.address.repo, self.address.path,
                            self.params['prefix'].lstrip('/'),
                            dst_base.lstrip('/'))

    @property
    def output_files(self):
        """Returns the list of output files from this rule.

        Paths are generated from the outputs of this rule's dependencies, with
        their paths translated based on prefix and strip_prefix.

        Returned paths are relative to buildroot.
        """
        for dep in self.subgraph.successors(self.address):
            dep_rule = self.subgraph.node[dep]['target_obj']
            for dep_file in dep_rule.output_files:
                yield self.translate_path(dep_file, dep_rule).lstrip('/')

    def composed_deps(self):
        for dep in self.params['srcs']:
            yield self.makeaddress(dep)

    @property
    def source_files(self):
        """pkgfilegroup doesn't have normal sources.

        This rule's srcs are actually handled as dependencies, which are then
        collected during the build phase.
        """
        return None



"""genrule targets"""

import os
import re
import subprocess
import sys
import stat
from butcher import address
from butcher import error
from butcher.targets import base
from twitter.common import log


class GenRuleBuilder(base.BaseBuilder):
    """Build a genrule."""

    # regex to match $(tags with parentheses) in cmd sublanguage
    paren_tag_re = re.compile(r'(?<!\$)\$\((.+?)\)')

    # regex to match $tags_without_parens in cmd sublanguage
    noparen_tag_re = re.compile(r'(?<!\$)\$([^\$.]+?)(?!\w)')

    # Shell to run cmd subprocesses in:
    shell_bin = '/bin/bash'

    def __init__(self, buildroot, target_obj, source_dir):
        base.BaseBuilder.__init__(self, buildroot, target_obj, source_dir)
        self.cmd = self.rule.params['cmd']
        self.path_to_this_rule = os.path.join(self.buildroot,
                                              self.rule.address.repo,
                                              self.rule.address.path)

    def build(self):
        log.debug('[%s]: Running in a shell:\n  %s', self.rule.name, self.cmd)
        cmd = [self.shell_bin, '-c', self.cmd]
        proc = subprocess.Popen(cmd, shell=False, cwd=self.path_to_this_rule,
                                stdout=sys.stdout, stderr=sys.stderr)
        returncode = proc.wait()
        if returncode != 0:
            raise error.TargetBuildFailed(self.rule.name,
                                          'cmd returned %s.' % (returncode))
        elif self.rule.params['executable']:
            # GenRule.__init__ already ensured that there is only one
            # output file if executable=1 is set.
            built_outfile = os.path.join(self.buildroot,
                                         self.rule.output_files[0])
            built_outfile_stat = os.stat(built_outfile)
            os.chmod(built_outfile, built_outfile_stat.st_mode | stat.S_IEXEC)

    def _metahash(self):
        """Include genrule cmd in the metahash."""
        if self._cached_metahash:
            return self._cached_metahash
        mhash = base.BaseBuilder._metahash(self)
        log.debug('[%s]: Metahash input: cmd="%s"', self.address, self.cmd)
        mhash.update(self.cmd)
        self._cached_metahash = mhash
        return mhash

    def prep(self):
        base.BaseBuilder.prep(self)
        self.expand_cmd_labels()

    def expand_cmd_labels(self):
        """Expand make-style variables in cmd parameters.

        Currently:
        $(location <foo>)     Location of one dependency or output file.
        $(locations <foo>)    Space-delimited list of foo's output files.
        $(SRCS)               Space-delimited list of this rule's source files.
        $(OUTS)               Space-delimited list of this rule's output files.
        $(@D)                 Full path to the output directory for this rule.
        $@                    Path to the output (single) file for this rule.
        """
        cmd = self.cmd

        def _expand_onesrc():
            """Expand $@ or $(@) to one output file."""
            outs = self.rule.params['outs'] or []
            if len(outs) != 1:
                raise error.TargetBuildFailed(
                    self.address,
                    '$@ substitution requires exactly one output file, but '
                    'this rule has %s of them: %s' % (len(outs), outs))
            else:
                return os.path.join(self.buildroot, self.path_to_this_rule,
                                    outs[0])

        # TODO: this function is dumb and way too long
        def _expand_makevar(re_match):
            """Expands one substitution symbol."""
            # Expand $(location foo) and $(locations foo):
            label = None
            tagstr = re_match.groups()[0]
            tag_location = re.match(
                r'\s*location\s+([A-Za-z0-9/\-_:\.]+)\s*', tagstr)
            tag_locations = re.match(
                r'\s*locations\s+([A-Za-z0-9/\-_:\.]+)\s*', tagstr)
            if tag_location:
                label = tag_location.groups()[0]
            elif tag_locations:
                label = tag_locations.groups()[0]
            if label:
                # Is it a filename found in the outputs of this rule?
                if label in self.rule.params['outs']:
                    return os.path.join(self.buildroot, self.address.repo,
                                        self.address.path, label)
                # Is it an address found in the deps of this rule?
                addr = self.rule.makeaddress(label)
                if addr not in self.rule.composed_deps():
                    raise error.TargetBuildFailed(
                        self.address,
                        '%s is referenced in cmd but is neither an output '
                        'file from this rule nor a dependency of this rule.' %
                        label)
                else:
                    paths = [x for x in self.rulefor(addr).output_files]
                    if len(paths) is 0:
                        raise error.TargetBuildFailed(
                            self.address,
                            'cmd refers to %s, but it has no output files.')
                    elif len(paths) > 1 and tag_location:
                        raise error.TargetBuildFailed(
                            self.address,
                            'Bad substitution in cmd: Expected exactly one '
                            'file, but %s expands to %s files.' % (
                                addr, len(paths)))
                    else:
                        return ' '.join(
                            [os.path.join(self.buildroot, x) for x in paths])

            # Expand $(OUTS):
            elif re.match(r'OUTS', tagstr):
                return ' '.join(
                    [os.path.join(self.buildroot, x)
                     for x in self.rule.output_files])

            # Expand $(SRCS):
            elif re.match(r'SRCS', tagstr):
                return ' '.join(os.path.join(self.path_to_this_rule, x)
                                for x in self.rule.params['srcs'] or [])

            # Expand $(@D):
            elif re.match(r'\s*@D\s*', tagstr):
                ruledir = os.path.join(self.buildroot, self.path_to_this_rule)
                return ruledir

            # Expand $(@), $@:
            elif re.match(r'\s*@\s*', tagstr):
                return _expand_onesrc()

            else:
                raise error.TargetBuildFailed(
                    self.address,
                    '[%s] Unrecognized substitution in cmd: %s' % (
                        self.address, re_match.group()))

        cmd, _ = re.subn(self.paren_tag_re, _expand_makevar, cmd)

        # Match tags starting with $ without parens. Will also catch parens, so
        # this goes after the tag_re substitutions.
        cmd, _ = re.subn(self.noparen_tag_re, _expand_makevar, cmd)

        # Now that we're done looking for $(blabla) and $bla parameters, clean
        # up any $$ escaping:
        cmd, _ = re.subn(r'\$\$', '$', cmd)

        # Maybe try heuristic label expansion?  Actually on second thought
        # that's a terrible idea. Use the explicit syntax, you lazy slobs. ;-)

        # TODO: Maybe consider other expansions from the gnu make manual?
        # $^ might be useful.
        # http://www.gnu.org/software/make/manual/html_node/Automatic-Variables.html#Automatic-Variables
        self.cmd = cmd


class GenRule(base.BaseTarget):
    """genrule target

    Arguments:
      name: A unique name for this rule. (required)
      srcs: List of inputs for this rule. (List of labels; optional)
      outs: List of output files generated by this command.
            (List of filenames; required)
      deps: List of dependencies for this rule. Will eventually be deprecated
            in favor of srcs and tools. (List of labels; optional)
      executable: If true, declares the output of this rule to be executable.
                  (Boolean; optional; default is False)
                  Setting this to True means the output is an executable file
                  and can be run using the "butcher run" command. The genrule
                  must produce exactly one output in this case.
      cmd: The command to run. (String; required)
           This argument is subject to some limited Make-style variable
           substitution:
           1. All occurrences of $(location <label>) are replaced by the full
              path to the file denoted by <label>. If <label> is malformed, or
              not a declared dependency of this rule, or does not expand to
              exactly one file, the rule will fail. The label does not need to
              be a canonical address: "foo", ":foo",
              "//fulladdr/to/module:rulename" are all valid.  The label may
              also be the name of an output file from this rule's "outs"
              attribute, in which case only the unqualified form (no // or :)
              may be used.
           2. All occurrences of $(locations <label>) are replaced by a
              space-separated list of paths to the files denoted by <label>.
              Similar to (1), the label must be a well-formed declared
              dependency of this rule, or one of its output files.
           3. $(OUTS) is replaced with a space-delimited list of full paths to
              this rule's output files. If the rule only has one output file,
              you may use $@ to refer to it.
           4. $(@D) is expanded to this rule's output directory.

    Notes:
      * $$ evaluates to $ (a literal dollar sign). For example, "$$$$.tmp" is a
        temporary filename containing the process ID of the shell ($$).
        Similarly, to invoke a shell command containing dollar-signs such as
        "ls $(dirname $x)", you should instead write "$((dirname $x)".
      * cmd should always use $(SRCS) or $(location <srclabel>) to refer to
        sources rather than hard-coding filenames. Likewise, it should use
        $(OUTS), $@, or $(location <outlabel>) to refer to output files. This
        is the only way to ensure that it reads and writes files in the right
        places. (outs does not write into the current directory, and neither
        should you: it may be read-only.)
      * The output directory is created automatically before the cmd runs.
      * In order to produce consistend and repeatible build results, you should
        avoid running binaries or depending on files that are not also built by
        butcher, or at the very least are checked into revision control and
        specified as srcs or deps of this rule. To run a binary built by
        butcher, it is best to refer to it by label, as its filename or
        location on disk may change.

    Examples:

      # The following rule generates a tarball from its srcs using gnu tar,
      # which is built from a hypothetical //tools:gnutar rule elsewhere:
      genrule(
          name = "docs",
          srcs = glob('docs/*.html') + glob('docs/*.png'),
          deps = ["//tools:gnutar"],
          outs = ["myfancydocumentation.tgz"],
          cmd = "$(location //tools:gnutar) czf $@ $(SRCS)",
          )
    """

    # TODO: srcs should really work with both filenames in source tree and
    #       labels referring to other rules. It's going to take some
    #       considerable refactoring of how "addresses" are handled in butcher
    #       to enable that, but I'm pretty sure that is the right way to go.
    # TODO: provide a way to declare user-defined "make" variables(?)
    # TODO: provide usage examples

    rulebuilder = GenRuleBuilder
    ruletype = 'genrule'

    required_params = [('name', str), ('cmd', str), ('outs', list)]
    optional_params = [('srcs', list, None),
                       ('deps', list, None),
                       ('executable', (int, bool), False)]

    def __init__(self, **kwargs):
        base.BaseTarget.__init__(self, **kwargs)

        if len(self.params['outs']) > 1 and self.params['executable']:
            raise error.InvalidRule(
                'executable=1 is only allowed when there is one output file.')

    @property
    def output_files(self):
        """Returns list of output files from this rule, relative to buildroot.

        In this case it's simple (for now) - the output files are enumerated in
        the rule definition.
        """
        outs = [os.path.join(self.address.repo, self.address.path, x)
                for x in self.params['outs']]
        return outs



"""filegroup targets"""

import os.path
from butcher.targets import base


class FileGroupBuilder(base.BaseBuilder):
    """Builder for filegroup."""
    ruletype = 'filegroup'

    def build(self):
        pass


class FileGroup(base.BaseTarget):
    """filegroup rule."""
    rulebuilder = FileGroupBuilder
    ruletype = 'filegroup'
    required_params = [('name', str), ('srcs', list)]
    optional_params = []

    @property
    def output_files(self):
        """Returns the list of output files from this rule.

        Paths are relative to buildroot.
        """
        for item in self.source_files:
            yield os.path.join(self.address.repo, self.address.path, item)



"""Manage rubygems for butcher's internal use.

Much like the rest of Butcher, this is fairly gross.
"""

import os
import subprocess
from twitter.common import app
from twitter.common import log
from butcher import error

app.add_option(
    '--gemdir', dest='gem_basedir', default='/var/lib/butcher',
    help='Path to our gems directory.')
app.add_option(
    '--gem_source', dest='gem_source', default='http://rubygems.org',
    help='Rubygems source repository.')

# TODO: Fallback gem_basedir for non-root installs?
# That is, most of the time butcher will come from a .deb that includes or
# depends on the requisite gems.  If it isn't, butcher should still be able to
# download and install what it needs in a user's homedir or elsewhere.
# Perhaps there should be a system /etc/butcherrc that can set things like
# gem_basedir in the case of distribution packages, and the default in this
# file could revert to being inside of the butcher work directory?


class RubyGems(app.Module):

    def __init__(self):
        app.Module.__init__(self, label=__name__,
                            description='Rubygems wrapper.',
                            dependencies='butcher')
        self.gem_basedir = None
        self.gem_source = None

    def gem_bindir(self):
        return os.path.join(self.gem_basedir, 'bin')

    def setup_function(self):
        if not app.get_options().gem_basedir:
            app.set_option('gem_basedir',
                           os.path.join(app.get_options().butcher_basedir,
                                        'gems'))
        self.gem_basedir = app.get_options().gem_basedir
        self.gem_source = app.get_options().gem_source
        os.environ['GEM_HOME'] = self.gem_basedir
        os.environ['GEM_PATH'] = self.gem_basedir

        try:
            subprocess.check_output(['gem', '--version'])
        except (OSError, subprocess.CalledProcessError):
            raise error.ButcherError('gem does not appear to be installed.')

app.register_module(RubyGems())


def install_gem(gemname, version=None, conservative=True, ri=False, rdoc=False,
                development=False, format_executable=False, force=False,
                gem_source=None):
    """Install a ruby gem."""
    cmdline = ['gem', 'install']
    if conservative:
        cmdline.append('--conservative')
    if ri:
        cmdline.append('--ri')
    else:
        cmdline.append('--no-ri')
    if rdoc:
        cmdline.append('--rdoc')
    else:
        cmdline.append('--no-rdoc')
    if development:
        cmdline.append('--development')
    if format_executable:
        cmdline.append('--format-executable')
    if force:
        cmdline.append('--force')
    if version:
        cmdline.extend(['--version', version])
    cmdline.extend(['--clear-sources',
                    '--source', gem_source or RubyGems().gem_source])

    cmdline.append(gemname)

    msg = 'Installing ruby gem: %s' % gemname
    if version:
        msg += ' Version requested: %s' % version
    log.debug(msg)

    try:
        subprocess.check_output(cmdline, shell=False)
    except (OSError, subprocess.CalledProcessError) as err:
        raise error.ButcherError(
            'Gem install failed. Error was: %s. Output: %s' % (
                err, err.output))


def is_installed(gemname, version=None):
    """Check if a gem is installed."""
    cmdline = ['gem', 'list', '-i', gemname]
    if version:
        cmdline.extend(['-v', version])
    try:
        subprocess.check_output(cmdline, shell=False)
        return True
    except (OSError, subprocess.CalledProcessError) as err:
        if err.returncode == 1:
            return False
        else:
            raise error.ButcherError(
                'Failure running gem. Error was: %s. Output: %s', err,
                err.output)


def gem_bindir():
    return RubyGems().gem_bindir()



"""Parse and handle BUILD files.

Each BUILD file is parsed into an acyclic digraph with each node representing
one build rule.

Two file formats are currently supported:
  1. JSON
  2. native Python

The latter is preferred and will be better supported as butcher gains features.
There are already a number of features and patterns that aren't easily
expressed in the JSON format, such as file globbing, variable substitution,
list comprehensions to format parameters, and custom functions for rule
generation.
"""

import json
import networkx
import os
import StringIO
from butcher import address
from butcher import buildfile_context
from butcher import error
from butcher import targets
from butcher import gitrepo
from twitter.common import app
from twitter.common import log


def load(stream, reponame, path):
    exception_list = []
    data = StringIO.StringIO(stream.read())
    for impl in (JsonBuildFile, PythonBuildFile):
        try:
            return impl(data, reponame, path)
        except ValueError as err:
            exception_list.append(err)
            data.seek(0)
    raise error.ButcherError('Unable to parse this buildfile: %s',
                             exception_list)


class BuildFile(networkx.DiGraph):
    """Base class for build file implementations."""

    def __init__(self, stream, reponame, path=''):
        self.target = address.new(repo=reponame, path=path)
        self.address = self.target
        #TODO: finish the target->address refactor
        networkx.DiGraph.__init__(self, name=self.target)

        # HACK ALERT: changing foreign module variables to set context...
        address.CUR_REPO = self.address.repo
        address.CUR_PATH = self.address.path

        try:
            self._parse(stream)
        except error.InvalidRule as err:
            raise error.BrokenGraph('%s (in file: %s)' % (
                err, os.path.join(self.path_on_disk,
                                  app.get_options().buildfile_name)))

        self.validate_internal_deps()

        # Add the :all node (unless it's explicitly defined in the build file)
        #                   (note: please don't do that to yourself)
        if self.target not in self.node:
            log.debug('New target: %s', self.target)
            self.add_node(
                self.target,
                {'target_obj': targets.new(name='all',
                                           ruletype='virtual',
                                           deps=[x for x in self.local_targets]
                                           )})

            for node in self.node:
                if node.repo == self.target.repo and node != self.target:
                    log.debug('New dep: %s -> %s', self.target, node)
                    self.add_edge(self.target, node)

        address.CUR_REPO = None
        address.CUR_PATH = ''

    def get_repo(self):
        return gitrepo.RepoState().GetRepo(self.address.repo)

    @property
    def path_on_disk(self):
        return os.path.join(self.get_repo().repo.tree().abspath,
                            self.address.path)

    def _parse(self, stream):
        """Parse and load a stream's contents into the digraph."""
        raise NotImplementedError  # Must override in subclass.

    def validate_internal_deps(self):
        """Freak out if there are missing local references."""
        for node in self.node:
            if ('target_obj' not in self.node[node]
                    and node not in self.crossrefs):
                raise error.BrokenGraph('Missing target: %s referenced from %s'
                                        ' but not defined there.' %
                                        (node, self.name))

    @property
    def crossrefs(self):
        """Returns a set of non-local targets referenced by this build file."""
        # TODO: memoize this?
        crefs = set()
        for node in self.node:
            if node.repo != self.target.repo or node.path != self.target.path:
                crefs.add(node)
        return crefs

    @property
    def crossref_paths(self):
        """Just like crossrefs, but all the targets are munged to :all."""
        return set(
            [address.new(repo=x.repo, path=x.path) for x in self.crossrefs])

    @property
    def local_targets(self):
        """Iterator over the targets defined in this build file."""
        for node in self.node:
            if (node.repo, node.path) == (self.target.repo, self.target.path):
                yield node


class JsonBuildFile(BuildFile):
    """JSON OCS_BUILD.data."""

    def __init__(self, stream, reponame, path=''):
        BuildFile.__init__(self, stream, reponame, path)

    def _parse(self, stream):
        """Parse a JSON BUILD file.

        Args:
          builddata: dictionary of buildfile data
          reponame: name of the repo that it came from
          path: directory path within the repo
        """
        builddata = json.load(stream)
        log.debug('This is a JSON build file.')

        if 'targets' not in builddata:
            log.warn('Warning: No targets defined here.')
            return

        for tdata in builddata['targets']:
            # TODO: validate name
            target = address.new(target=tdata.pop('name'),
                                 repo=self.target.repo,
                                 path=self.target.path)
            # Duplicate target definition? Uh oh.
            if target in self.node and 'target_obj' in self.node[target]:
                raise error.ButcherError(
                    'Target is defined more than once: %s', target)

            rule_obj = targets.new(name=target,
                                   ruletype=tdata.pop('type'),
                                   **tdata)

            log.debug('New target: %s', target)
            self.add_node(target, {'target_obj': rule_obj})

            # dep could be ":blabla" or "//foo:blabla" or "//foo/bar:blabla"
            for dep in rule_obj.composed_deps() or []:
                d_target = address.new(dep)
                if not d_target.repo:  # ":blabla"
                    d_target.repo = self.target.repo
                if d_target.repo == self.target.repo and not d_target.path:
                    d_target.path = self.target.path
                if d_target not in self.nodes():
                    self.add_node(d_target)
                log.debug('New dep: %s -> %s', target, d_target)
                self.add_edge(target, d_target)


class PythonBuildFile(BuildFile):
    """Python-based build file implementation!"""

    def __init__(self, stream, reponame, path=''):
        self.code = None
        BuildFile.__init__(self, stream, reponame, path)

    def _parse(self, stream):
        self.code = compile(stream.read(), str(self.target), 'exec')
        targets.base.BaseTarget.graphcontext = self
        context = buildfile_context.ParseContext(self)
        context.parse()

        for node in self.nodes():
            for dep in self.node[node]['target_obj'].composed_deps() or []:
                d_target = address.new(dep)
                if not d_target.repo:
                    d_target.repo = self.target.repo
                if d_target.repo == self.target.repo and not d_target.path:
                    d_target.path = self.target.path
                if d_target not in self.nodes():
                    self.add_node(d_target)
                log.debug('New dep: %s -> %s', node, d_target)
                self.add_edge(node, d_target)
